@@ -66,7 +66,7 @@ ARGV   *mail_addr_crunch(const char *string, const char *extension)
     TOK822 **addr_list;
     TOK822 **tpp;
     char   *ratsign;
-    int     extlen;
+    ssize_t extlen;
 
     if (extension)
 	extlen = strlen(extension);
@@ -77,7 +77,12 @@ ARGV   *mail_addr_crunch(const char *string, const char *extension)
      * Parse the string, rewrite each address to canonical form, and convert
      * the result to external (quoted) form. Optionally apply the extension
      * to each address found.
+     * 
+     * XXX Workaround for the null address. This works for envelopes but
+     * produces ugly results for message headers.
      */
+    if (*string == 0 || strcmp(string, "<>") == 0)
+	string = "\"\"";
     tree = tok822_parse(string);
     addr_list = tok822_grep(tree, TOK822_ADDR);
     for (tpp = addr_list; *tpp; tpp++) {
@@ -108,6 +113,7 @@ ARGV   *mail_addr_crunch(const char *string, const char *extension)
  /*
   * Stand-alone test program, sort of interactive.
   */
+#include <stdlib.h>
 #include <unistd.h>
 #include <msg.h>
 #include <vstream.h>
@@ -139,6 +145,7 @@ int     main(int unused_argc, char **unused_argv)
 	    vstream_printf("	%s\n", *cpp);
 	vstream_fflush(VSTREAM_OUT);
     }
+    return (0);
 }
 
 #endif
