@@ -39,6 +39,7 @@
 /*	int	var_dont_remove;
 /*	char	*var_inet_interfaces;
 /*	char	*var_proxy_interfaces;
+/*	char	*var_inet_protocols;
 /*	char	*var_mynetworks;
 /*	char	*var_double_bounce_sender;
 /*	int	var_line_limit;
@@ -103,8 +104,12 @@
 /*	int     var_strict_encoding;
 /*	int     var_verify_neg_cache;
 /*	int	var_oldlog_compat;
+/*	int	var_delay_max_res;
+/*	char	*var_int_filt_classes;
 /*
 /*	void	mail_params_init()
+/*
+/*	const	char null_format_string[1];
 /* DESCRIPTION
 /*	This module (actually the associated include file) define the names
 /*	and defaults of all mail configuration parameters.
@@ -114,6 +119,9 @@
 /*	initialized globally so as to avoid hard-to-find errors due to
 /*	missing initialization. This routine must be called early, at
 /*	least before entering a chroot jail.
+/*
+/*	null_format_string is a workaround for gcc compilers that complain
+/*	about empty or null format strings.
 /* DIAGNOSTICS
 /*	Fatal errors: out of memory; null system or domain name.
 /* LICENSE
@@ -152,15 +160,17 @@
 #ifdef HAS_DB
 #include <dict_db.h>
 #endif
+#include <inet_proto.h>
 
 /* Global library. */
 
-#include "mynetworks.h"
-#include "mail_conf.h"
-#include "mail_version.h"
-#include "mail_proto.h"
-#include "verp_sender.h"
-#include "mail_params.h"
+#include <mynetworks.h>
+#include <mail_conf.h>
+#include <mail_version.h>
+#include <mail_proto.h>
+#include <verp_sender.h>
+#include <own_inet_addr.h>
+#include <mail_params.h>
 
  /*
   * Special configuration variables.
@@ -198,6 +208,7 @@ char   *var_pid_dir;
 int     var_dont_remove;
 char   *var_inet_interfaces;
 char   *var_proxy_interfaces;
+char   *var_inet_protocols;
 char   *var_mynetworks;
 char   *var_double_bounce_sender;
 int     var_line_limit;
@@ -262,6 +273,10 @@ int     var_strict_8bit_body;
 int     var_strict_encoding;
 int     var_verify_neg_cache;
 int     var_oldlog_compat;
+int     var_delay_max_res;
+char   *var_int_filt_classes;
+
+const char null_format_string[1] = "";
 
 /* check_myhostname - lookup hostname and validate */
 
@@ -279,16 +294,13 @@ static const char *check_myhostname(void)
 
     /*
      * If the local machine name is not in FQDN form, try to append the
-     * contents of $mydomain.
+     * contents of $mydomain. Use a default domain as a final workaround.
      */
     name = get_hostname();
     if ((dot = strchr(name, '.')) == 0) {
 	if ((domain = mail_conf_lookup_eval(VAR_MYDOMAIN)) == 0)
-	    msg_warn("My hostname %s is not a fully qualified name - set %s or %s in %s/%s",
-		     name, VAR_MYHOSTNAME, VAR_MYDOMAIN,
-		     var_config_dir, MAIN_CONF_FILE);
-	else
-	    name = concatenate(name, ".", domain, (char *) 0);
+	    domain = DEF_MYDOMAIN;
+	name = concatenate(name, ".", domain, (char *) 0);
     }
     return (name);
 }
@@ -300,11 +312,10 @@ static const char *check_mydomainname(void)
     char   *dot;
 
     /*
-     * Use the hostname when it is not a FQDN ("foo"), or when the hostname
-     * actually is a domain name ("foo.com").
+     * Use a default domain when the hostname is not a FQDN ("foo").
      */
-    if ((dot = strchr(var_myhostname, '.')) == 0 || strchr(dot + 1, '.') == 0)
-	return (var_myhostname);
+    if ((dot = strchr(var_myhostname, '.')) == 0)
+	return (DEF_MYDOMAIN);
     return (dot + 1);
 }
 
@@ -428,6 +439,7 @@ void    mail_params_init()
 {
     static CONFIG_STR_TABLE first_str_defaults[] = {
 	VAR_SYSLOG_FACILITY, DEF_SYSLOG_FACILITY, &var_syslog_facility, 1, 0,
+	VAR_INET_PROTOCOLS, DEF_INET_PROTOCOLS, &var_inet_protocols, 1, 0,
 	0,
     };
     static CONFIG_STR_FN_TABLE function_str_defaults[] = {
@@ -478,6 +490,7 @@ void    mail_params_init()
 	VAR_FLUSH_SERVICE, DEF_FLUSH_SERVICE, &var_flush_service, 1, 0,
 	VAR_VERIFY_SERVICE, DEF_VERIFY_SERVICE, &var_verify_service, 1, 0,
 	VAR_TRACE_SERVICE, DEF_TRACE_SERVICE, &var_trace_service, 1, 0,
+	VAR_INT_FILT_CLASSES, DEF_INT_FILT_CLASSES, &var_int_filt_classes, 0, 0,
 	0,
     };
     static CONFIG_STR_FN_TABLE function_str_defaults_2[] = {
@@ -500,6 +513,7 @@ void    mail_params_init()
 	VAR_TOKEN_LIMIT, DEF_TOKEN_LIMIT, &var_token_limit, 1, 0,
 	VAR_MIME_MAXDEPTH, DEF_MIME_MAXDEPTH, &var_mime_maxdepth, 1, 0,
 	VAR_MIME_BOUND_LEN, DEF_MIME_BOUND_LEN, &var_mime_bound_len, 1, 0,
+	VAR_DELAY_MAX_RES, DEF_DELAY_MAX_RES, &var_delay_max_res, MIN_DELAY_MAX_RES, MAX_DELAY_MAX_RES,
 	0,
     };
     static CONFIG_TIME_TABLE time_defaults[] = {
@@ -532,6 +546,7 @@ void    mail_params_init()
 	0,
     };
     const char *cp;
+    INET_PROTO_INFO *proto_info;
 
     /*
      * Extract syslog_facility early, so that from here on all errors are
@@ -544,6 +559,12 @@ void    mail_params_init()
 		  var_config_dir, MAIN_CONF_FILE,
 		  VAR_SYSLOG_FACILITY, var_syslog_facility);
 
+    /*
+     * What protocols should we attempt to support? The result is stored in
+     * the global inet_proto_table variable.
+     */
+    proto_info = inet_proto_init(VAR_INET_PROTOCOLS, var_inet_protocols);
+
     /*
      * Variables whose defaults are determined at runtime. Some sites use
      * short hostnames in the host table; some sites name their system after
@@ -582,6 +603,13 @@ void    mail_params_init()
      */
     get_mail_conf_str_fn_table(function_str_defaults_2);
 
+    /*
+     * FIX 200412 The IPv6 patch did not call own_inet_addr_list() before
+     * entering the chroot jail on Linux IPv6 systems. Linux has the IPv6
+     * interface list in /proc, which is not available after chrooting.
+     */
+    (void) own_inet_addr_list();
+
     /*
      * The PID variable cannot be set from the configuration file!!
      */
