@@ -20,9 +20,9 @@
 /*	QMGR_TRANSPORT *transport;
 /*	const char *name;
 /*
-/*	void	qmgr_queue_throttle(queue, reason)
+/*	void	qmgr_queue_throttle(queue, dsn)
 /*	QMGR_QUEUE *queue;
-/*	const char *reason;
+/*	DSN	*dsn;
 /*
 /*	void	qmgr_queue_unthrottle(queue)
 /*	QMGR_QUEUE *queue;
@@ -120,7 +120,7 @@ static void qmgr_queue_unthrottle_wrapper(int unused_event, char *context)
 
 void    qmgr_queue_unthrottle(QMGR_QUEUE *queue)
 {
-    char   *myname = "qmgr_queue_unthrottle";
+    const char *myname = "qmgr_queue_unthrottle";
     QMGR_TRANSPORT *transport = queue->transport;
 
     if (msg_verbose)
@@ -131,10 +131,10 @@ void    qmgr_queue_unthrottle(QMGR_QUEUE *queue)
      */
     if (queue->window == 0) {
 	event_cancel_timer(qmgr_queue_unthrottle_wrapper, (char *) queue);
-	if (queue->reason == 0)
-	    msg_panic("%s: queue %s: window 0 reason 0", myname, queue->name);
-	myfree(queue->reason);
-	queue->reason = 0;
+	if (queue->dsn == 0)
+	    msg_panic("%s: queue %s: window 0 status 0", myname, queue->name);
+	dsn_free(queue->dsn);
+	queue->dsn = 0;
 	queue->window = transport->init_dest_concurrency;
 	return;
     }
@@ -152,18 +152,19 @@ void    qmgr_queue_unthrottle(QMGR_QUEUE *queue)
 
 /* qmgr_queue_throttle - handle destination delivery failure */
 
-void    qmgr_queue_throttle(QMGR_QUEUE *queue, const char *reason)
+void    qmgr_queue_throttle(QMGR_QUEUE *queue, DSN *dsn)
 {
-    char   *myname = "qmgr_queue_throttle";
+    const char *myname = "qmgr_queue_throttle";
 
     /*
      * Sanity checks.
      */
-    if (queue->reason)
+    if (queue->dsn)
 	msg_panic("%s: queue %s: spurious reason %s",
-		  myname, queue->name, queue->reason);
+		  myname, queue->name, queue->dsn->reason);
     if (msg_verbose)
-	msg_info("%s: queue %s: %s", myname, queue->name, reason);
+	msg_info("%s: queue %s: %s %s",
+		 myname, queue->name, dsn->status, dsn->reason);
 
     /*
      * Decrease the destination's concurrency limit until we reach zero, at
@@ -178,9 +179,10 @@ void    qmgr_queue_throttle(QMGR_QUEUE *queue, const char *reason)
      * Special case for a site that just was declared dead.
      */
     if (queue->window == 0) {
-	queue->reason = mystrdup(reason);
+	queue->dsn = DSN_COPY(dsn);
 	event_request_timer(qmgr_queue_unthrottle_wrapper,
 			    (char *) queue, var_min_backoff_time);
+	queue->dflags = 0;
     }
 }
 
@@ -188,7 +190,7 @@ void    qmgr_queue_throttle(QMGR_QUEUE *queue, const char *reason)
 
 void    qmgr_queue_done(QMGR_QUEUE *queue)
 {
-    char   *myname = "qmgr_queue_done";
+    const char *myname = "qmgr_queue_done";
     QMGR_TRANSPORT *transport = queue->transport;
 
     /*
@@ -202,9 +204,9 @@ void    qmgr_queue_done(QMGR_QUEUE *queue)
 	msg_panic("%s: queue not empty: %s", myname, queue->name);
     if (queue->window <= 0)
 	msg_panic("%s: window %d", myname, queue->window);
-    if (queue->reason)
+    if (queue->dsn)
 	msg_panic("%s: queue %s: spurious reason %s",
-		  myname, queue->name, queue->reason);
+		  myname, queue->name, queue->dsn->reason);
 
     /*
      * Clean up this in-core queue.
@@ -231,6 +233,8 @@ QMGR_QUEUE *qmgr_queue_create(QMGR_TRANSPORT *transport, const char *name,
 
     queue = (QMGR_QUEUE *) mymalloc(sizeof(QMGR_QUEUE));
     qmgr_queue_count++;
+    queue->dflags = 0;
+    queue->last_done = 0;
     queue->name = mystrdup(name);
     queue->nexthop = mystrdup(nexthop);
     queue->todo_refcount = 0;
@@ -239,7 +243,7 @@ QMGR_QUEUE *qmgr_queue_create(QMGR_TRANSPORT *transport, const char *name,
     queue->window = transport->init_dest_concurrency;
     QMGR_LIST_INIT(queue->todo);
     QMGR_LIST_INIT(queue->busy);
-    queue->reason = 0;
+    queue->dsn = 0;
     queue->clog_time_to_warn = 0;
     queue->blocker_tag = 0;
     QMGR_LIST_APPEND(transport->queue_list, queue, peers);
