@@ -11,7 +11,7 @@
  /*
   * System library.
   */
-#include <time.h>
+#include <sys/time.h>
 
  /*
   * Utility library.
@@ -32,10 +32,12 @@ typedef struct {
     VSTREAM *client;			/* client connection */
     VSTRING *message;			/* message buffer */
     VSTRING *buf;			/* line buffer */
-    time_t  time;			/* start of session */
+    struct timeval arrival_time;	/* start of session */
     char   *name;			/* client name */
     char   *addr;			/* client IP address */
     char   *namaddr;			/* name[addr] */
+    char   *rfc_addr;			/* RFC 2821 client IP address */
+    int     addr_family;		/* address family */
     char   *queue_id;			/* queue file ID */
     VSTREAM *cleanup;			/* cleanup server */
     MAIL_STREAM *dest;			/* cleanup server */
