@@ -95,7 +95,8 @@
 /*	with the MAIL_CONFIG environment variable to override the
 /*	location of configuration files.
 /* .IP "\fB-F \fIfull_name\fR
-/*	Set the sender full name. This is used only with messages that
+/*	Set the sender full name. This overrides the NAME environment
+/*	variable, and is used only with messages that
 /*	have no \fBFrom:\fR message header.
 /* .IP "\fB-f \fIsender\fR"
 /*	Set the envelope sender address. This is the address where
@@ -227,11 +228,15 @@
 /* .fi
 /* .IP \fBMAIL_CONFIG\fR
 /*	Directory with Postfix configuration files.
-/* .IP \fBMAIL_VERBOSE\fR
+/* .IP "\fBMAIL_VERBOSE\fR (value does not matter)"
 /*	Enable verbose logging for debugging purposes.
-/* .IP \fBMAIL_DEBUG\fR
+/* .IP "\fBMAIL_DEBUG\fR (value does not matter)"
 /*	Enable debugging with an external command, as specified with the
 /*	\fBdebugger_command\fR configuration parameter.
+/* .IP \fBNAME\fR
+/*	The sender full name. This is used only with messages that
+/*	have no \fBFrom:\fR message header. See also the \fB-F\fR
+/*	option above.
 /* CONFIGURATION PARAMETERS
 /* .ad
 /* .fi
@@ -338,9 +343,6 @@
 /* .IP "\fBsyslog_name (postfix)\fR"
 /*	The mail system name that is prepended to the process name in syslog
 /*	records, so that "smtpd" becomes, for example, "postfix/smtpd".
-/* .IP "\fBtrigger_timeout (10s)\fR"
-/*	The time limit for sending a trigger to a Postfix daemon (for
-/*	example, the \fBpickup\fR(8) or \fBqmgr\fR(8) daemon).
 /* FILES
 /*	/var/spool/postfix, mail queue
 /*	/etc/postfix, configuration files
@@ -883,6 +885,13 @@ static void enqueue(const int flags, const char *encoding,
     myfree(saved_sender);
 }
 
+/* tempfail - sanitize exit status after library run-time error */
+
+static void tempfail(void)
+{
+    exit(EX_TEMPFAIL);
+}
+
 /* main - the main program */
 
 int     main(int argc, char **argv)
@@ -950,6 +959,7 @@ int     main(int argc, char **argv)
     if ((slash = strrchr(argv[0], '/')) != 0 && slash[1])
 	argv[0] = slash + 1;
     msg_vstream_init(argv[0], VSTREAM_ERR);
+    msg_cleanup(tempfail);
     msg_syslog_init(mail_task("sendmail"), LOG_PID, LOG_FACILITY);
     set_mail_conf_str(VAR_PROCNAME, var_procname = mystrdup(argv[0]));
 
