@@ -230,6 +230,11 @@ char   *xfer_request[SMTP_STATE_LAST] = {
     "QUIT command",
 };
 
+#define SMTP_MIME_DOWNGRADE(session, request) \
+    (var_disable_mime_oconv == 0 \
+     && (session->features & SMTP_FEATURE_8BITMIME) == 0 \
+     && strcmp(request->encoding, MAIL_ATTR_ENC_7BIT) != 0)
+
 static int smtp_start_tls(SMTP_STATE *);
 
 /* smtp_helo - perform initial handshake with SMTP server */
@@ -1009,6 +1014,8 @@ static int smtp_loop(SMTP_STATE *state, NOCLOBBER int send_state,
     } while (0)
 
 #define RETURN(x) do { \
+	if (recv_state != SMTP_STATE_LAST) \
+	    DONT_CACHE_THIS_SESSION; \
 	vstring_free(next_command); \
 	if (survivors) \
 	    myfree((char *) survivors); \
@@ -1172,7 +1179,9 @@ static int smtp_loop(SMTP_STATE *state, NOCLOBBER int send_state,
 	    QUOTE_ADDRESS(session->scratch, vstring_str(session->scratch2));
 	    vstring_sprintf(next_command, "MAIL FROM:<%s>",
 			    vstring_str(session->scratch));
-	    if (session->features & SMTP_FEATURE_SIZE)	/* RFC 1870 */
+	    /* XXX Don't announce SIZE if we're going to MIME downgrade. */
+	    if (session->features & SMTP_FEATURE_SIZE	/* RFC 1870 */
+		&& !SMTP_MIME_DOWNGRADE(session, request))
 		vstring_sprintf_append(next_command, " SIZE=%lu",
 				       request->data_size);
 	    if (session->features & SMTP_FEATURE_8BITMIME) {	/* RFC 1652 */
@@ -1619,13 +1628,13 @@ static int smtp_loop(SMTP_STATE *state, NOCLOBBER int send_state,
 	 * transaction in progress.
 	 */
 	if (send_state == SMTP_STATE_DOT && nrcpt > 0) {
-	    downgrading =
-		(var_disable_mime_oconv == 0
-		 && (session->features & SMTP_FEATURE_8BITMIME) == 0
-		 && strcmp(request->encoding, MAIL_ATTR_ENC_7BIT) != 0);
+	    downgrading = SMTP_MIME_DOWNGRADE(session, request);
+	    /* XXX Don't downgrade just because generic_maps is turned on. */
 	    if (downgrading || smtp_generic_maps)
-		session->mime_state = mime_state_alloc(MIME_OPT_DOWNGRADE
-						  | MIME_OPT_REPORT_NESTING,
+		session->mime_state = mime_state_alloc(downgrading ?
+						       MIME_OPT_DOWNGRADE
+						 | MIME_OPT_REPORT_NESTING :
+						       MIME_OPT_DISABLE_MIME,
 						       smtp_generic_maps ?
 						       smtp_header_rewrite :
 						       smtp_header_out,
