@@ -54,11 +54,12 @@
 
 int     stream_send_fd(int fd, int sendfd)
 {
-    char   *myname = "stream_send_fd";
+    const char *myname = "stream_send_fd";
 
 #ifdef STREAM_CONNECTIONS
     if (ioctl(fd, I_SENDFD, sendfd) < 0)
 	msg_fatal("%s: send file descriptor: %m", myname);
+    return (0);
 #else
     msg_fatal("stream connections are not implemented");
 #endif
@@ -72,6 +73,8 @@ int     stream_send_fd(int fd, int sendfd)
   */
 #include <unistd.h>
 #include <fcntl.h>
+#include <string.h>
+#include <stdlib.h>
 #include <split_at.h>
 #include <connect.h>
 
@@ -104,7 +107,7 @@ int     main(int argc, char **argv)
 	if (stream_send_fd(server_sock, client_fd) < 0)
 	    msg_fatal("send file descriptor: %m");
 	if (close(client_fd) != 0)
-	    msg_fatal("close(%d): %m");
+	    msg_fatal("close(%d): %m", client_fd);
     }
     exit(0);
 }
