@@ -28,6 +28,7 @@
     || defined(BSDI2) || defined(BSDI3) || defined(BSDI4) \
     || defined(OPENBSD2) || defined(OPENBSD3) || defined(OPENBSD4) \
     || defined(NETBSD1) || defined(NETBSD2) || defined(NETBSD3) \
+    || defined(NETBSD4) \
     || defined(EKKOBSD1)
 #define SUPPORTED
 #include <sys/types.h>
@@ -512,6 +513,7 @@ extern int opterr;
 # define HAS_IPV6
 #endif
 #define BROKEN_AI_PASSIVE_NULL_HOST
+#define BROKEN_AI_NULL_SERVICE
 #endif
 
 #ifdef AIX4
