@@ -122,7 +122,7 @@ static WATCHDOG *watchdog_curr;
 
 static void watchdog_event(int unused_sig)
 {
-    char   *myname = "watchdog_event";
+    const char *myname = "watchdog_event";
     WATCHDOG *wp;
 
     /*
@@ -148,7 +148,7 @@ static void watchdog_event(int unused_sig)
 
 WATCHDOG *watchdog_create(unsigned timeout, WATCHDOG_FN action, char *context)
 {
-    char   *myname = "watchdog_create";
+    const char *myname = "watchdog_create";
     struct sigaction sig_action;
     WATCHDOG *wp;
 
@@ -177,7 +177,7 @@ WATCHDOG *watchdog_create(unsigned timeout, WATCHDOG_FN action, char *context)
 
 void    watchdog_destroy(WATCHDOG *wp)
 {
-    char   *myname = "watchdog_destroy";
+    const char *myname = "watchdog_destroy";
 
     watchdog_stop(wp);
     watchdog_curr = wp->saved_watchdog;
@@ -194,7 +194,7 @@ void    watchdog_destroy(WATCHDOG *wp)
 
 void    watchdog_start(WATCHDOG *wp)
 {
-    char   *myname = "watchdog_start";
+    const char *myname = "watchdog_start";
 
     if (wp != watchdog_curr)
 	msg_panic("%s: wrong watchdog instance", myname);
@@ -208,7 +208,7 @@ void    watchdog_start(WATCHDOG *wp)
 
 void    watchdog_stop(WATCHDOG *wp)
 {
-    char   *myname = "watchdog_stop";
+    const char *myname = "watchdog_stop";
 
     if (wp != watchdog_curr)
 	msg_panic("%s: wrong watchdog instance", myname);
@@ -221,7 +221,7 @@ void    watchdog_stop(WATCHDOG *wp)
 
 void    watchdog_pat(void)
 {
-    char   *myname = "watchdog_pat";
+    const char *myname = "watchdog_pat";
 
     if (watchdog_curr)
 	watchdog_curr->trip_run = 0;
@@ -233,7 +233,7 @@ void    watchdog_pat(void)
 
 #include <vstream.h>
 
-main(int unused_argc, char **unused_argv)
+int     main(int unused_argc, char **unused_argv)
 {
     WATCHDOG *wp;
 
@@ -245,6 +245,7 @@ main(int unused_argc, char **unused_argv)
 	watchdog_pat();
     } while (VSTREAM_GETCHAR() != VSTREAM_EOF);
     watchdog_destroy(wp);
+    return (0);
 }
 
 #endif
